# -*- coding: utf-8 -*-
#
# satextrato/__init__.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
from __future__ import unicode_literals

import logging


__version__ = '0.3'

logging.getLogger('satextrato').addHandler(logging.NullHandler())


from .venda import ExtratoCFeVenda  # noqa: E402, F401
from .cancelamento import ExtratoCFeCancelamento  # noqa: E402, F401
